/* Calculate the likelihood scores of the domain family pairs in an input file.
   Downloadable program.
   Inputs: one or two families of aligned sequences, one or two gene trees, species tree, substitution rate matrix, prior, penality parameter, filtering parameters, llratio threshold.
   Outputs: position pairs passing the threshold.
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "db.h"
# define MAXLINEWIDTH 100000
# define NAMELENGTH 100
# define LONGNAMELENGTH 100
# define PHYLIPWIDTH 10
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern int inlist(int Nv, int *v, int target);
extern char * getitemval2(char *s, int ind, int *length);
extern char * getitemval3(char *s, int ind, int *length, char sepch);
extern double atof2(char *s, int length);
extern char *i2a(long num, int *length);
extern int nu2ind(char ch);
extern int aa2ind(char ch);
extern void parsetree(char *treestring, int nseq, char *names, int *child_pnt, int *t_pnt, int *order_pnt);
extern double * kron(int nrows1, int ncols1, double *mat1, int nrows2, int ncols2, double *mat2);
extern void calculate_joint_rate(double *q1, double *q2, double *q2d);
extern void padm(int ndim, double *A, double t, int p, double *E);
extern void get_sequence_tree(int nseq1, int lseq1, char *accs1, char *seqs1, int *gene2sp1, int ngenetreenodes1, int **genetreechild1, double *t1, int nseq2, int lseq2, char *accs2, char *seqs2, int *gene2sp2, int ngenetreenodes2, int **genetreechild2, double *t2, int nspecies, char *speciesnames, int nsptreenodes, int **sptreechild, int *retnseq, int *subaccs1_pnt, int *subseqs1_pnt, int *subaccs2_pnt, int *subseqs2_pnt, int *t_pnt);
extern double llratio(int nnodes, int **child, double *t, int *order, int nseq, int *seqs, double *prior2, double *nulltransitions, double *activetransitions);
extern double llratio_quantized_rates(int nnodes, int **child, double *t, int *order, int nseq, int *seqs, int Nbllevels, double *bllevels, double *prior2, double *nulltransitions, double *activetransitions);

int nchar, njointchar, nnodes;
double gapthre, consthre, covthre, llthre, epsilon;
char *aamap=" ARNDCQEGHILKMFPSTWYV";
char *numap=" ACGU";

int main(int argc, char *argv[])
{
  // *(argv+1) -- 0:RNA, 1:protein.
  // *(argv+2) -- intra or inter molecules, intra: 0, inter: 1.
  // *(argv+3) -- sequence family 1.
  // *(argv+4) -- sequence family 2 (same as family 1 for intra molecules).
  // *(argv+5) -- gene tree 1.
  // *(argv+6) -- gene tree 2 (same as gene tree 1 for intra molecules).
  // *(argv+7) -- species names.
  // *(argv+8) -- species tree.
  // *(argv+9) -- sequence 1 name to species name mapping.
  // *(argv+10) -- sequence 2 name to species name mapping.
  // *(argv+11) -- prior of single nucleotide or amino acid.
  // *(argv+12) -- rate matrix of single nucleotide or amino acid.
  // *(argv+13) -- penalty parameter (epsilon).
  // *(argv+14) -- gap threshold.
  // *(argv+15) -- conservation threshold.
  // *(argv+16) -- covariation threshold.
  // *(argv+17) -- llratio threshold.
  // *(argv+18) -- position pairs and scores that pass the threshold.
  // *(argv+19) -- mode of calculation, 0:scan the entire sequences, 1:calculate a pair of positions.
  // *(argv+20) -- selected genes for each family, valid only when inter=1.  
  // *(argv+21) -- the position 1 for single pair calculation, valid only when mode=1.
  // *(argv+22) -- the position 2 for single pair calculation, valid only when mode=1.

  char ch, *s, *item, *filename, *treestring, *seqs1, *seqs2, *accs1, *accs2, *speciesnames;
  int i, j, k, l, m, n, s1, s2, length, flag, inter, molecule, *order, *order1, *order2, mode;
  int pnt1, pnt2, pnt3, pnt4, pnt5, pnt6, pnt7, pnt8;
  int nseq, nseq1, nseq2, lseq1, lseq2, pos1, pos2, *pairs, npairs;
  int ngenetreenodes1, ngenetreenodes2, **genetreechild1, **genetreechild2;
  int nsptreenodes, **sptreechild, nspecies, *gene2sp1, *gene2sp2;
  double *t1, *t2, *t, tmp, *prior, *prior2, *q1, *q2, *q2d, *activetrans, *nulltrans, *scores;  
  double gapthre, consthre, covthre, llthre, score, *bllevels;
  int *pairseqs, jcnts[21][21], phylipwidth, Nbllevels;
  FILE *fp, *fout;
  
  // Load indicator for RNA or protein.
  molecule=atoi(*(argv+1));

  if (molecule==0)
    {
      nchar=4; njointchar=16;
    }
  else if (molecule==1)
    {
      nchar=20; njointchar=400;
    }
  else
    {
      printf("Please choose 0 (RNA) or 1 (protein) for the first argument.\n");
      return -1;
    }

  // Load indicator for intra or inter molecules.
  inter=atoi(*(argv+2));
  if ((inter!=0)&&(inter!=1))
    {
      printf("Please choose 0 (intra molecule) or 1 (inter molecule) for the second argument.\n");
      return -1;
    }

  s=malloc(sizeof(char)*MAXLINEWIDTH);

  
  // Import sequence family 1 in phylip format.
  if ((fp=fopen(*(argv+3),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+3));
      return -1;
    }  

  // Get the number of sequences and sequence length.
  fgets(s,MAXLINEWIDTH,fp);
  item=getitemval2(s,0,&length); tmp=atof2(item,length);
  nseq1=(int)tmp; free(item);
  item=getitemval2(s,1,&length); tmp=atof2(item,length);
  lseq1=(int)tmp; free(item);

  // Get the sequences and names.
  fgets(s,MAXLINEWIDTH,fp); phylipwidth=0; i=0;
  while ((ch=*(s+i))!=' ')
    i++;
  while ((ch=*(s+i))==' ')
    i++;
  phylipwidth=i; fseek(fp,0,SEEK_SET); fgets(s,MAXLINEWIDTH,fp); 
  seqs1=malloc(sizeof(char)*nseq1*lseq1); accs1=malloc(sizeof(char)*NAMELENGTH*nseq1);
  memset(seqs1,'-',nseq1*lseq1); i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      j=0;
      while ((ch=*(s+j))!=' ')
	j++;
      strncpy(accs1+i*NAMELENGTH,s,j);
      *(accs1+i*NAMELENGTH+j)='\0';
      for (j=0; j<=(lseq1-1); j++)
	{
	  ch=*(s+j+phylipwidth);
	  *(seqs1+i*lseq1+j)=ch;
	}
      i++;     
    }
  fclose(fp);

  
  // Debug
  //printf("nseq1=%d, lseq1=%d\n",nseq1,lseq1);
  

  // If inter=1 get the sequences and names of family 2.
  if (inter==1)
    {
      if ((fp=fopen(*(argv+4),"r"))==NULL)
	{
	  printf("Cannot open %s.\n",*(argv+4));
	  return -1;
	}  

      // Get the number of sequences and sequence length.
      fgets(s,MAXLINEWIDTH,fp);
      item=getitemval2(s,0,&length); tmp=atof2(item,length);
      nseq2=(int)tmp; free(item);
      item=getitemval2(s,1,&length); tmp=atof2(item,length);
      lseq2=(int)tmp; free(item);
      
      // Get the sequences and names.
      fgets(s,MAXLINEWIDTH,fp); phylipwidth=0; i=0;
      while ((ch=*(s+i))!=' ')
	i++;
      while ((ch=*(s+i))==' ')
	i++;
      phylipwidth=i; fseek(fp,0,SEEK_SET); fgets(s,MAXLINEWIDTH,fp); 
      seqs2=malloc(sizeof(char)*nseq2*lseq2); accs2=malloc(sizeof(char)*NAMELENGTH*nseq2);
      memset(seqs2,'-',nseq2*lseq2); i=0;
      while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
	{
	  j=0;
	  while ((ch=*(s+j))!=' ')
	    j++;
	  strncpy(accs2+i*NAMELENGTH,s,j);
          *(accs2+i*NAMELENGTH+j)='\0';
	  for (j=0; j<=(lseq2-1); j++)
	    {
	      ch=*(s+j+phylipwidth);
	      *(seqs2+i*lseq2+j)=ch;
	    }
	  i++;     
	}
      fclose(fp);

      // Debug
      //printf("nseq2=%d, lseq2=%d\n",nseq2,lseq2);
        
    }
  
  // Debug
  printf("Load and parse gene trees.\n");

  // Import and parse the gene tree of family 1.
  if ((fp=fopen(*(argv+5),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+5));
      return -1;
    }
  length=0; treestring=malloc(sizeof(char)*MAXLINEWIDTH);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      i=0;
      while ((ch=*(s+i))!=0x0a)
	{
	  *(treestring+length)=ch; i++; length++;
	}
    }
  *(treestring+length)='\0'; fclose(fp);
  nnodes=0; 
  parsetree(treestring,nseq1,accs1,&pnt1,&pnt2,&pnt3);
  genetreechild1=pnt1; t1=pnt2; order1=pnt3; ngenetreenodes1=nnodes; 
  
  // Debug
  //printf("ngenetreenodes1=%d, nseq1=%d\n",ngenetreenodes1,nseq1);   

  // If inter=1 import and parse the gene tree of family 2.
  if (inter==1)
    {
      if ((fp=fopen(*(argv+6),"r"))==NULL)
	{
	  printf("Cannot open %s.\n",*(argv+6));
	  return -1;
	}
      length=0; 
      while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
	{
	  i=0;
	  while ((ch=*(s+i))!=0x0a)
	    {
	      *(treestring+length)=ch; i++; length++;
	    }
	}
      *(treestring+length)='\0'; fclose(fp);
      
      nnodes=0; 
      parsetree(treestring,nseq2,accs2,&pnt1,&pnt2,&pnt3);
      genetreechild2=pnt1; t2=pnt2; order2=pnt3; ngenetreenodes2=nnodes;

      // Debug
      //printf("ngenetreenodes2=%d, nseq2=%d\n",ngenetreenodes2,nseq2);  
      
    }
  else
    {
      genetreechild2=genetreechild1; t2=t1; order2=order1; ngenetreenodes2=ngenetreenodes1;
    }

  // Debug
  printf("Load and pass the species tree.\n");  

  // Load the species names.
  // Each species name is in a separate line.
  nspecies=0; speciesnames=malloc(sizeof(char)*NAMELENGTH);
  if ((fp=fopen(*(argv+7),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+7));
      return -1;
    }
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      nspecies++; speciesnames=realloc(speciesnames,sizeof(char)*nspecies*NAMELENGTH);
      i=0;
      while ((ch=*(s+i))!=0x0a)
	i++;
      strncpy(speciesnames+(nspecies-1)*NAMELENGTH,s,i);
      *(speciesnames+(nspecies-1)*NAMELENGTH+i)='\0';
    }
  fclose(fp);
  
  // Debug
  //printf("nspecies=%d\n",nspecies);    

  // Import and parse the species tree.
  if ((fp=fopen(*(argv+8),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+8));
      return -1;
    }
  length=0; 
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      i=0;
      while ((ch=*(s+i))!=0x0a)
	{
	  *(treestring+length)=ch; i++; length++;
	}
    }
  *(treestring+length)='\0'; fclose(fp);
  nnodes=0; 
  parsetree(treestring,nspecies,speciesnames,&pnt1,&pnt2,&pnt3);
  nsptreenodes=nnodes;
  sptreechild=pnt1; t=pnt2; order=pnt3; nsptreenodes=nnodes; free(t); 
  
  // Debug
  //printf("nsptreenodes=%d, nspecies=%d\n",nsptreenodes,nspecies);  

  // Load the mapping from sequence family 1 to species.
  gene2sp1=malloc(sizeof(int)*nseq1);
  for (i=0; i<=(nseq1-1); i++)
    *(gene2sp1+i)=-1;
  if ((fp=fopen(*(argv+9),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+9));
      return -1;
    }
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval3(s,0,&length,0x09);
      k=0; i=-1;
      while ((k<nseq1)&&(i==-1))
	{
	  if (strncmp(accs1+k*NAMELENGTH,item,max(strlen(accs1+k*NAMELENGTH),length))==0)
	    i=k;
	  k++;
	}
      free(item);
      item=getitemval3(s,1,&length,0x09);
      k=0; j=-1;
      while ((k<nseq1)&&(j==-1))
	{
	  if (strncmp(speciesnames+k*NAMELENGTH,item,max(strlen(speciesnames+k*NAMELENGTH),length))==0)
	    j=k;
	  k++;
	}
      free(item);      
      if ((i>=0)&&(j>=0))
	*(gene2sp1+i)=j;
    }
  fclose(fp);

  
  // If inter=1, load the mapping from sequence family 2 to species.
  if (inter==1)
    {
      gene2sp2=malloc(sizeof(int)*nseq2);
      for (i=0; i<=(nseq2-1); i++)
	*(gene2sp2+i)=-1;
      if ((fp=fopen(*(argv+10),"r"))==NULL)
	{
	  printf("Cannot open %s.\n",*(argv+10));
	  return -1;
	}
      while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
	{
	  item=getitemval3(s,0,&length,0x09);
	  k=0; i=-1;
	  while ((k<nseq2)&&(i==-1))
	    {
	      if (strncmp(accs2+k*NAMELENGTH,item,max(strlen(accs2+k*NAMELENGTH),length))==0)
		i=k;
	      k++;
	    }
	  free(item);
	  item=getitemval3(s,1,&length,0x09);
	  k=0; j=-1;
	  while ((k<nseq2)&&(j==-1))
	    {
	      if (strncmp(speciesnames+k*NAMELENGTH,item,max(strlen(speciesnames+k*NAMELENGTH),length))==0)
		j=k;
	      k++;
	    }
	  free(item);
	  
	  if ((i>=0)&&(j>=0))
	    *(gene2sp2+i)=j;
	}
      fclose(fp);
     
    }    

  // Debug
  printf("Load parameters for the CTMP model.\n");
  
  // Load the prior of single nucleotides or amino acids.
  prior=malloc(sizeof(double)*nchar); 
  if ((fp=fopen(*(argv+11),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+11));
      return -1;
    }
  fgets(s,MAXLINEWIDTH,fp);
  for (i=0; i<=(nchar-1); i++)
    {
      item=getitemval2(s,i,&length);      
      *(prior+i)=atof2(item,length);
      free(item);
    }
  fclose(fp); 
  
  // Load the rate matrix of single nucleotides or amino acids.
  q1=malloc(sizeof(double)*nchar*nchar);
  if ((fp=fopen(*(argv+12),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+12));
      return -1;
    }
  i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      for (j=0; j<=(nchar-1); j++)
	{
	  item=getitemval2(s,j,&length);
	  *(q1+i*nchar+j)=atof2(item,length);
	  free(item);
	}
      i++;
    }
  fclose(fp);
  
  // Load the penalty parameter epsilon.
  epsilon=atof(*(argv+13));

  // Calculate the joint prior and independent rate matrix.
  prior2=kron(1,nchar,prior,1,nchar,prior);
  q2=malloc(sizeof(double)*njointchar*njointchar);
  q2d=malloc(sizeof(double)*njointchar*njointchar);
  calculate_joint_rate(q1,q2,q2d);
   
  
  // If inter=1, get the joint tree and aligned sequences of the two families.
  if (inter==1)
    {
      // Debug
      printf("Identify representatives from the two families.\n");

      get_sequence_tree(nseq1,lseq1,accs1,seqs1,gene2sp1,ngenetreenodes1,genetreechild1,t1,nseq2,lseq2,accs2,seqs2,gene2sp2,ngenetreenodes2,genetreechild2,t2,nspecies,speciesnames,nsptreenodes,sptreechild,&nseq,&pnt1,&pnt2,&pnt3,&pnt4,&pnt5);
      free(accs1); free(seqs1); free(accs2); free(seqs2);       
      accs1=pnt1; seqs1=pnt2; accs2=pnt3; seqs2=pnt4; 
      t=pnt5;

      // Report the selected genes in each family.
      if ((fout=fopen(*(argv+20),"w"))==NULL)
	{
	  printf("Cannot open %s.\n",*(argv+20));
	  return -1;
	}
      fprintf(fout,"species%cgene 1%cgene 2\n",0x09,0x09);
      for (i=0; i<=(nspecies-1); i++)
	fprintf(fout,"%s%c%s%c%s\n",speciesnames+i*LONGNAMELENGTH,0x09,accs1+i*NAMELENGTH,0x09,accs2+i*NAMELENGTH);
      fclose(fout);

    }

  // If inter=0, copy the information of the first family to the second one.
  if (inter==0)
    {
      lseq2=lseq1; nseq2=nseq1; seqs2=seqs1; nseq=nseq2; 
      nsptreenodes=ngenetreenodes1; sptreechild=genetreechild1; t=t1; order=order1;
    }
  
  // Debug
  printf("Pre-compute transition probabilities along each branch.\n");
  
  // RNA: calculate the exact conditional probability along the branch.
  if (molecule==0)
    {
      activetrans=malloc(sizeof(double)*njointchar*njointchar*nsptreenodes);
      nulltrans=malloc(sizeof(double)*njointchar*njointchar*nsptreenodes);  
      for (i=0; i<=(nsptreenodes-1); i++)
	{      
	  padm(njointchar,q2d,*(t+i),7,activetrans+i*njointchar*njointchar);
	  padm(njointchar,q2,*(t+i),7,nulltrans+i*njointchar*njointchar);
	}
    }

  // Protein: pre-compute the quantized transition probabilities.  Cannot compute and store the exact conditional probabilities due to the large space required.
  else if (molecule==1)
    {
      Nbllevels=10; bllevels=malloc(sizeof(double)*Nbllevels);  
      *(bllevels+0)=0.0; *(bllevels+1)=0.01; *(bllevels+2)=0.02; *(bllevels+3)=0.05;
      *(bllevels+4)=0.08; *(bllevels+5)=0.1; *(bllevels+6)=0.2; *(bllevels+7)=0.5;
      *(bllevels+8)=0.8; *(bllevels+9)=1.0;
      nulltrans=malloc(sizeof(double)*njointchar*njointchar*Nbllevels);
      activetrans=malloc(sizeof(double)*njointchar*njointchar*Nbllevels);
  
      for (i=0; i<=(Nbllevels-1); i++)
	{
	  padm(njointchar,q2,*(bllevels+i)*0.01,6,nulltrans+i*njointchar*njointchar);
	  padm(njointchar,q2d,*(bllevels+i)*0.01,6,activetrans+i*njointchar*njointchar);
	}     
    }
  

  // Load the thresholds for screening and reporting the coevolving pairs.
  gapthre=atof(*(argv+14)); consthre=atof(*(argv+15)); 
  covthre=atof(*(argv+16)); llthre=atof(*(argv+17)); 
      
  // Debug
  printf("Calculate the coevolutionary scores.\n");

  // Launch the screening on single or double families of sequences.
  // Applying filtering on the joint sequences.         
  mode=atoi(*(argv+19));
  if (mode==1)
    {
      pos1=atoi(*(argv+21)); pos2=atoi(*(argv+22)); pos1-=1; pos2-=1;
    }
  pairseqs=malloc(sizeof(int)*nseq*2); 
  npairs=0; pairs=malloc(sizeof(int)*2); scores=malloc(sizeof(double));
  for (i=0; i<=(lseq1-2); i++)  
    {      
      for (j=0; j<=(lseq2-1); j++)      
	{	  
	  // For self interactions, only consider j>i.
	  // If mode=1, then only consider pair (pos1,pos2).
	  if (((inter==1)||((inter==0)&&(j>i)))&&((mode==0)||((mode==1)&&(i==pos1)&&(j==pos2))))
	    {	      
	      for (k=0; k<=(nseq-1); k++)
		{
		  ch=*(seqs1+k*lseq1+i); 
		  if (nchar==4)
		    *(pairseqs+k*2+0)=nu2ind(ch);
		  else
		    *(pairseqs+k*2+0)=aa2ind(ch);
		  ch=*(seqs2+k*lseq2+j); 
		  if (nchar==4)
		    *(pairseqs+k*2+1)=nu2ind(ch);
		  else
		    *(pairseqs+k*2+1)=aa2ind(ch);
		}
	     
	      // Apply gap filtering.
	      flag=1; k=0;
	      for (l=0; l<=(nseq-1); l++)
		{	
		  if (*(pairseqs+l*2+0)==0)
		    k++;		 	 
		}

	      if ((double)k>=(double)(nseq*gapthre))
		flag=0;
	      if (flag==1)
		{
		  k=0;
		  for (l=0; l<=(nseq-1); l++)
		    {
		      if (*(pairseqs+l*2+1)==0)
			k++;		      
		    }
		 
		  if ((double)k>=(double)(nseq*gapthre))
		    flag=0;
		}

	      // Apply gap filtering on joint states.
	      if (flag==1)
		{		  
		  for (k=0; k<=nchar; k++)
		    for (l=0; l<=nchar; l++)
		      jcnts[k][l]=0;
		  
		  for (k=0; k<=(nseq-1); k++)
		    {
		      l=*(pairseqs+k*2+0);
		      m=*(pairseqs+k*2+1);
		      jcnts[l][m]+=1;		      
		    }
		 
		  l=jcnts[0][0];		 
		  for (k=1; k<=nchar; k++)		
		    l+=(jcnts[k][0]+jcnts[0][k]);

		  if (l>=(gapthre*nseq))
		    flag=0;		 
		}
	      
	      // Apply conservation filtering.
	      if (flag==1)
		{
		  int maxcnts=0;
		  for (k=0; k<=nchar; k++)
		    for (l=0; l<=nchar; l++)
		      if (jcnts[k][l]>maxcnts)
			maxcnts=jcnts[k][l];
		  if (maxcnts>=(consthre*nseq))
		    flag=0;		
		}

	      // Apply covariation filtering.
	      if (flag==1)
		{
		  int nstates, *states1, *states2, occcnts, flag2;

		  flag2=1; nstates=0; occcnts=0;
		  states1=malloc(sizeof(int)*nchar); states2=malloc(sizeof(int)*nchar);
		  while (flag2==1)
		    {
		      // Find the joint state which does not share row or column with reported states and has the highest occurrences.
		      m=0; s1=-1; s2=-1;
		      for (k=1; k<=nchar; k++)
			{
			  for (l=1; l<=nchar; l++)
			    {
			      if ((inlist(nstates,states1,k)==0)&&(inlist(nstates,states2,l)==0)&&(jcnts[k][l]>m))
				{
				  s1=k; s2=l; m=jcnts[k][l];
				}
			    }
			}
		      
		      // If no such states exist then stop.
		      if (m<=0)
			flag2=0;
		      
		      // Otherwise incorporate the new state.
		      else
			{
			  nstates++; *(states1+nstates-1)=s1; *(states2+nstates-1)=s2; occcnts+=m;				      			  
			}
		    }

		  if (occcnts<(nseq*covthre))
		    flag=0;
		  free(states1); free(states2);
		}
	    
	      // Compute the likelihood score.
	      if (flag==1)
		{
		  // RNA.
		  if (molecule==0)
		    score=llratio(nsptreenodes,sptreechild,t,order,nseq,pairseqs,prior2,nulltrans,activetrans);
		  // Protein.
		  else if (molecule==1)
		    score=llratio_quantized_rates(nsptreenodes,sptreechild,t,order,nseq,pairseqs,Nbllevels,bllevels,prior2,nulltrans,activetrans);
		  
		  if (score>=llthre)
		    {
		      npairs++; pairs=realloc(pairs,sizeof(int)*npairs*2);
		      scores=realloc(scores,sizeof(double)*npairs);
		      *(pairs+(npairs-1)*2+0)=i;
		      *(pairs+(npairs-1)*2+1)=j;
		      *(scores+npairs-1)=score;
		    }		  
		}
	    }
	}
    }

  // Report the pairs that pass the threshold.
  if ((fout=fopen(*(argv+18),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+18));
      return -1;
    }
  fprintf(fout,"position 1%cposition 2%cscore\n",0x09,0x09);
  for (i=0; i<=(npairs-1); i++)
    fprintf(fout,"%d%c%d%c%f\n",*(pairs+i*2+0)+1,0x09,*(pairs+i*2+1)+1,0x09,*(scores+i));
  fclose(fout);

  return 0;
}


